/*
 * Copyright 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include "BilinearLessBlur.h"
#include "DColor.h"
#include <cstring>
#include <cmath>

bool BilinearLessBlur::needStretchBuffer() const
{
	return false;
}

void BilinearLessBlur::renderStretch(const SDL_Rect& srcArea, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	int bpp = destination->format->BytesPerPixel;
	switch (bpp)
	{
	case 3:
		renderStretchT<Scale24>(srcArea, dstArea, source, destination);
		break;
	case 4:
		renderStretchT<Scale32>(srcArea, dstArea, source, destination);
		break;
	}
}

void BilinearLessBlur::renderZoom(const SDL_Rect& srcArea, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	renderStretch(srcArea, dstArea, source, destination);
}

template <typename S>
void BilinearLessBlur::renderStretchT(const SDL_Rect& srcArea_, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	// This code is very similar to the one in Bilinear.cpp. To get
	// less blur the distances are squared and then scaled so that
	// the sums add up to 1.
	
	typedef typename S::PixelType PixelType;
	
	SDL_Rect srcArea = srcArea_;
	if (srcArea.x > 0)
	{
		--srcArea.x;
		++srcArea.w;
	}
	if (srcArea.y > 0)
	{
		--srcArea.y;
		++srcArea.h;
	}
	
	int xOffset = xZoom / 2;
	int yOffset = yZoom / 2;
	
	bool leftBorder = srcArea.x == 0;
	bool rightBorder = srcArea.x + srcArea.w == GAME_WIDTH;
	bool topBorder = srcArea.y == 0;
	bool bottomBorder = srcArea.y + srcArea.h == GAME_HEIGHT;
	
	dstArea.x = src2dst_x(srcArea.x);
	dstArea.y = src2dst_y(srcArea.y);
	dstArea.w = src2dst_x(srcArea.x + srcArea.w) - dstArea.x;
	dstArea.h = src2dst_y(srcArea.y + srcArea.h) - dstArea.y;
	
	dstArea.x += xOffset;
	dstArea.y += yOffset;
	
	if (leftBorder)
	{
		dstArea.x -= xOffset;
		dstArea.w += xOffset;
	}
	if (topBorder)
	{
		dstArea.y -= yOffset;
		dstArea.h += yOffset;
	}
	if (rightBorder)
	{
		dstArea.w -= xOffset;
	}
	else
	{
		// counteract rounding errors
		++dstArea.w;
	}
	if (bottomBorder)
	{
		dstArea.h -= yOffset;
	}
	else
	{
		// counteract rounding errors
		++dstArea.h;
	}
	
	DColor c;
	DColor c1;
	DColor c2;
	DColor c3;
	DColor c4;
	
	int xStart = dstArea.x;
	int yStart = dstArea.y;
	int xEnd = dstArea.x + dstArea.w;
	int yEnd = dstArea.y + dstArea.h;
	
	double xStep = static_cast<double>(GAME_WIDTH) / screenArea.w;
	double yStep = static_cast<double>(GAME_HEIGHT) / screenArea.h;
	
	double xDistanceStart = (xStart - xOffset - screenArea.x) * xStep;
	int srcXStart = std::floor(xDistanceStart);
	xDistanceStart = xDistanceStart - srcXStart;
	
	double _yDistance1 = (yStart - yOffset - screenArea.y) * yStep;
	int srcY = std::floor(_yDistance1);
	_yDistance1 = _yDistance1 - srcY;
	
	PixelType* p1;
	PixelType* p2;
	PixelType* p3;
	PixelType* p4;
	
	PixelType* src = (PixelType*) source->pixels + S::pixelSize * srcXStart + srcY * source->pitch / sizeof(PixelType);
	PixelType* dst = (PixelType*) destination->pixels + S::pixelSize * dstArea.x + dstArea.y * destination->pitch / sizeof(PixelType);

	std::ptrdiff_t src_rowskip = source->pitch / sizeof(PixelType);
	std::ptrdiff_t dst_rowskip = destination->pitch / sizeof(PixelType);

	for (int y = yStart; y < yEnd; ++y)
	{
		int srcX = srcXStart;
		double _xDistance1 = xDistanceStart;
		p1 = src;
		p2 = src + S::pixelSize;
		p3 = src + src_rowskip;
		p4 = src + src_rowskip + S::pixelSize;
		
		if (srcX < 0)
		{
			p1 = p2;
			p3 = p4;
		}
		if (srcY < 0)
		{
			p1 = p3;
			p2 = p4;
		}
		else if (srcY >= GAME_HEIGHT - 1)
		{
			p3 = p1;
			p4 = p2;
		}
		
		S::readColors(p1, c1);
		S::readColors(p2, c2);
		S::readColors(p3, c3);
		S::readColors(p4, c4);
		
		double _yDistance2 = 1.0 - _yDistance1;
		double yDistance1 = _yDistance1 * _yDistance1;
		double yDistance2 = _yDistance2 * _yDistance2;
		double k = (yDistance1 + yDistance2);
		yDistance1 /= k;
		yDistance2 /= k;
		PixelType *dstTmp = dst;
		
		double ry1 = c1.r * yDistance2 + c3.r * yDistance1;
		double ry2 = c2.r * yDistance2 + c4.r * yDistance1;
		double gy1 = c1.g * yDistance2 + c3.g * yDistance1;
		double gy2 = c2.g * yDistance2 + c4.g * yDistance1;
		double by1 = c1.b * yDistance2 + c3.b * yDistance1;
		double by2 = c2.b * yDistance2 + c4.b * yDistance1;
		
		for (int x = xStart; x < xEnd; ++x)
		{
			if (_xDistance1 >= 1.0)
			{
				_xDistance1 -= 1.0;
				//p1 = p2;
				c1 = c2;
				//p3 = p4;
				c3 = c4;
				++srcX;
				if (srcX < GAME_WIDTH - 1)
				{
					p2 += S::pixelSize;
					p4 += S::pixelSize;
					S::readColors(p2, c2);
					S::readColors(p4, c4);
				}
				ry1 = c1.r * yDistance2 + c3.r * yDistance1;
				ry2 = c2.r * yDistance2 + c4.r * yDistance1;
				gy1 = c1.g * yDistance2 + c3.g * yDistance1;
				gy2 = c2.g * yDistance2 + c4.g * yDistance1;
				by1 = c1.b * yDistance2 + c3.b * yDistance1;
				by2 = c2.b * yDistance2 + c4.b * yDistance1;
			}
			double _xDistance2 = 1.0 - _xDistance1;
			double xDistance1 = _xDistance1 * _xDistance1;
			double xDistance2 = _xDistance2 * _xDistance2;
			double k = (xDistance1 + xDistance2);
			xDistance1 /= k;
			xDistance2 /= k;
			
			c.r = ry1 * xDistance2 + ry2 * xDistance1;
			c.g = gy1 * xDistance2 + gy2 * xDistance1;
			c.b = by1 * xDistance2 + by2 * xDistance1;
			
			S::writeColors(dst, c);
			
			dst += S::pixelSize;
			_xDistance1 += xStep;
		}
		_yDistance1 += yStep;
		if (_yDistance1 >= 1.0)
		{
			_yDistance1 -= 1.0;
			src += src_rowskip;
			++srcY;
		}
		dst = dstTmp + dst_rowskip;
	}
	
}
